@@ -13,6 +13,7 @@ from learning_loop_node.node import Node, State
 import logging
 from datetime import datetime
 from ..socket_response import SocketResponse
+from .helper import is_valid_uuid4
 
 
 class TrainerNode(Node):
@@ -68,13 +69,20 @@ class TrainerNode(Node):
             self.trainer.stop_training()
             await self.update_state(State.Idle)
             return
-        self.latest_known_model_id = source_model['id']
+        # NOTE 'id' can also be a well known pretrained model identifier.
+        self.latest_known_model_id = source_model['id'] if is_valid_uuid4(source_model['id']) else None
         await self.update_state(State.Running)
 
-    async def stop_training(self) -> Union[bool, str]:
+    async def stop_training(self, save_latest_model: bool = True) -> Union[bool, str]:
         self.status.reset_error('stop_training')
         try:
             result = self.trainer.stop_training()
+
+            # NOTE saving the without having an own new checkpoint does not harm.
+            if save_latest_model and self.latest_known_model_id and self.trainer.source_model_id != self.latest_known_model_id:
+                await self.save_model(self.trainer.training.context, self.latest_known_model_id)
+            await self.clear_training_data(self.trainer.training.training_folder)
+
             self.trainer.training = None
             self.latest_known_model_id = None
 
@@ -100,14 +108,27 @@ class TrainerNode(Node):
 
         await self.send_status()
 
+    async def clear_training_data(self, training_folder: str):
+        self.status.reset_error('clear_training_data')
+        try:
+            await self.trainer.clear_training_data(training_folder)
+        except Exception as e:
+            traceback.print_exc()
+            self.status.set_error('clear_training_data', f'Could not delete training data: {str(e)}')
+
     async def check_state(self):
         logging.debug(f'{self.status.state}')
         self.status.reset_error('training_error')
         error = self.trainer.get_error()
 
         if error is not None:
-            logging.error(error + '\n\n' + self.trainer.get_log()[-1000:])
+            try:
+                # NOTE test_model_should_be_uploaded_when_training_has_error will result in exception without this try/except block.
+                logging.error(error + '\n\n' + self.trainer.get_log()[-1000:])
+            except:
+                pass
             self.status.set_error('training_error', error)
+            await self.stop_training()
             await self.send_status()
             return
 
@@ -117,6 +138,7 @@ class TrainerNode(Node):
         if not self.trainer.executor.is_process_running():
             self.status.set_error('training_error', 'Training crashed.')
             logging.info(self.trainer.get_log()[-1000:])
+            await self.stop_training()
             await self.send_status()
             return
 
@@ -148,7 +170,7 @@ class TrainerNode(Node):
                         logging.error(error_msg)
                         raise Exception(error_msg)
 
-                    logging.info(f'successfully uploaded model {jsonable_encoder(new_model)}')
+                    logging.info(f'successfully uploaded checkpoint {jsonable_encoder(new_model)}')
                     self.trainer.on_model_published(model, new_model.id)
                     self.latest_known_model_id = new_model.id
 
@@ -169,6 +191,9 @@ class TrainerNode(Node):
             latest_produced_model_id=self.latest_known_model_id,
             errors=self.status._errors
         )
+        # TODO can self.trainer be None?
+        if self.trainer:
+            status.pretrained_models = self.trainer.provided_pretrained_models
 
         if self.trainer and self.trainer.training:
             status.train_image_count = self.trainer.training.data.train_image_count()
@@ -181,6 +206,9 @@ class TrainerNode(Node):
 
         if not response.success:
             logging.error(f'Error for updating: Response from loop was : {response.__dict__}')
+            logging.error('Going to kill training. ')
+            if status.state != State.Idle:
+                await self.stop_training(save_latest_model=False)
 
     def get_state(self):
         if self.trainer.executor is not None and self.trainer.executor.is_process_running():
